/* ////////////////////////////////////////////////////////////////////////////
 * File:        SyResPtr.h
 *
 * Purpose:     Resource pointer class implementation.
 *
 * Created:     24th November 2000
 * Updated:     6th October 2005
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Public License)
 *
 *              Copyright 2000-2005, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au
 *                          http://www.synesis.com.au/software
 *
 *              email:      software@synesis.com.au
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Redistributions in binary form must reproduce the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer in the documentation
 *              and/or other materials provided with the distribution.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iv) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (v) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef _SYNSOFT_INCL_H_SYRESPTR
#define _SYNSOFT_INCL_H_SYRESPTR

#ifndef _SYNSOFT_DOCUMENTATION_SKIP_SECTION
# define _SYNSOFT_VER_H_SYRESPTR_MAJOR        3
# define _SYNSOFT_VER_H_SYRESPTR_MINOR        3
# define _SYNSOFT_VER_H_SYRESPTR_REVISION     3
# define _SYNSOFT_VER_H_SYRESPTR_EDIT         614
#endif /* !_SYNSOFT_DOCUMENTATION_SKIP_SECTION */

////////////////////////////////////////////////////////////////////////////////
// Includes

#include <MLCpp.h>
#ifdef _SYB_TRACE_INCLUDE_PATH
#pragma message(__FILE__)
#endif /* _SYB_TRACE_INCLUDE_PATH */

////////////////////////////////////////////////////////////////////////////////
// Classes

#ifdef __SYNSOFT_DBS_TEMPLATE_SUPPORT

////////////////////////////////////////////////////////////////////////////////

template <typename_arg_t T>
struct IsNull;

template <typename_arg_t T>
struct SetNull;

template <typename_arg_t T>
struct IsNull_generator
{
    typedef IsNull<T>   type;
};

////////////////////////////////////////////////////////////////////////////////
// class Resource_ptr
//
// The Resource_ptr class provides a (semi-smart) pointer wrapper implementation
// in order to ensure auotmatic release/destruction/discard of a resource.
//
// The Resource_ptr class is paramaterised by the following:
//
//  T           -   type of resource, e.g. void *
//  TC          -   const type of resource, which is often not simply const T,
//                  e.g. void const *
//  FnIsNull    -   a functor (functional) type whose function operator returns
//                  a Bool value indicating whether the given resource handle
//                  is "null". The semantic of the "null" value depends on the
//                  resource, but is generally equivalent to 0.
//  FnClose     -   a functor (functional) type whose function operator releases
//                  the given resource.
//  FnSetNull   -   a functor (functional) type whose function operator sets the
//                  given reference to its "null" value (see FnIsNull).
//

template<   typename_arg_t T
        ,   typename_arg_t TC
        ,   typename_arg_t FnIsNull
        ,   typename_arg_t FnClose
        ,   typename_arg_t FnSetNull
        >
class Resource_ptr
{
public:
    typedef T                                                   TResource;
    typedef TC                                                  TCResource;
    typedef Resource_ptr<T, TC, FnIsNull, FnClose, FnSetNull>   Class;
    typedef _mR(Class)                                          ClassRef;
    typedef _mRC(Class)                                         ClassConstRef;

public:
    Resource_ptr()
    {
        FnSetNull()(m_t);
    }
    /* explicit  */Resource_ptr(TResource t)
        : m_t(t)
    {}
    Resource_ptr(ClassRef rhs)
        : m_t(rhs.m_t)
    {
        FnSetNull()(rhs.m_t);
    }
    ~Resource_ptr() SyThrow_None()
    {
        if(!FnIsNull()(m_t))
        {
            FnClose()(m_t);
        }
    }

    ClassRef operator =(TResource t)
    {
        if(!FnIsNull()(m_t))
        {
            FnClose()(m_t);
        }

        m_t = t;

        return *this;
    }

    ClassRef operator =(Class &rhs)
    {
        if(&rhs != this)
        {
            if(!FnIsNull()(m_t))
            {
                FnClose()(m_t);
            }

            m_t = rhs.m_t;
            FnSetNull()(rhs.m_t);
        }

        return *this;
    }

    TResource Detach()
    {
        TResource   t = m_t;

        FnSetNull()(m_t);

        return t;
    }
    void Close()
    {
        if(!FnIsNull()(m_t))
        {
            FnClose()(m_t);
            FnSetNull()(m_t);
        }
    }

    __SYNSOFT_GEN_NS_QUAL(::SynesisStd, Bool) IsNull() const
    {
        return FnIsNull()(m_t);
    }

    operator TCResource () const
    {
        return m_t;
    }

    operator TResource ()
    {
        return m_t;
    }

protected:
    TResource   m_t;
};

// class SimpleResource_ptr

template<   typename_arg_t T
        ,   typename_arg_t FnClose
        >
class SimpleResource_ptr
    : public Resource_ptr<T, T, IsNull<T>, FnClose, SetNull<T> >
{
public:
    typedef T                                                   TResource;
private:
    typedef Resource_ptr<   T
                        ,   T
                        ,   typename_type_t IsNull_generator<T>::type
                        ,   FnClose
                        ,   SetNull<T>
                        >                                       parent_class_type;
public:
#if __SYNSOFT_DVS_COMPILER_VER_MSVC == __SYNSOFT_GEN_COMPILER_VERSION(7, 1)
    typedef parent_class_type                                   ParentClass;
#else /* ? __SYNSOFT_DVS_COMPILER_VER_MSVC != __SYNSOFT_GEN_COMPILER_VERSION(7, 1) */
    typedef Resource_ptr<T, T, IsNull<T>, FnClose, SetNull<T> > ParentClass;
#endif /* __SYNSOFT_DVS_COMPILER_VER_MSVC != __SYNSOFT_GEN_COMPILER_VERSION(7, 1) */
    typedef SimpleResource_ptr<T, FnClose>                      Class;
    typedef _mR(Class)                                          ClassRef;
    typedef _mRC(Class)                                         ClassConstRef;

public:
    SimpleResource_ptr()
    {}
    /* explicit  */SimpleResource_ptr(TResource t)
        : ParentClass(t)
    {}
    SimpleResource_ptr(ClassRef rhs)
        : ParentClass(rhs)
    {}

    ClassRef operator =(TResource t)
    {
        ParentClass::operator =(t);

        return *this;
    }

    ClassRef operator =(Class &rhs)
    {
        ParentClass::operator =(rhs);

        return *this;
    }

};

////////////////////////////////////////////////////////////////////////////////
// Shims

#if __SYNSOFT_DVS_COMPILER != __SYNSOFT_VAL_COMPILER_DMC
template<   typename_arg_t T
        ,   typename_arg_t TC
        ,   typename_arg_t FnIsNull
        ,   typename_arg_t FnClose
        ,   typename_arg_t FnSetNull
        >
inline __SYNSOFT_GEN_NS_QUAL(::SynesisStd, Bool) is_empty(Resource_ptr<T, TC, FnIsNull, FnClose, FnSetNull> const &p)
{
    return p.IsNull();
}

template<   typename_arg_t T
        ,   typename_arg_t TC
        ,   typename_arg_t FnIsNull
        ,   typename_arg_t FnClose
        ,   typename_arg_t FnSetNull
        >
inline typename_type_t Resource_ptr<T, TC, FnIsNull, FnClose, FnSetNull>::TCResource get_ptr(Resource_ptr<T, TC, FnIsNull, FnClose, FnSetNull> const &p)
{
    return p.operator TC();
}
#endif /* compiler */

////////////////////////////////////////////////////////////////////////////////

template <typename_arg_t T>
struct IsNull
{
    __SYNSOFT_GEN_NS_QUAL(::SynesisStd, Bool) operator ()(_mRC(T) t)
    {
        return t == NULL;
    }
};

template <typename_arg_t T>
struct SetNull
{
    void operator ()(_mR(T) t)
    {
        t = NULL;
    }
};

#endif // __SYNSOFT_DBS_TEMPLATE_SUPPORT

////////////////////////////////////////////////////////////////////////////////

#endif // _SYNSOFT_INCL_H_SYRESPTR

////////////////////////////////////////////////////////////////////////////////
