/* /////////////////////////////////////////////////////////////////////////////
 * File:        MLTssStr.cpp
 *
 * Purpose:     Synesis Software Intelligent-TSS API.
 *
 * Created:     21st January 1999
 * Updated:     22nd October 2005
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Private Supply License)
 *
 *              Copyright 1999-2005, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au
 *                          http://www.synesis.com.au/software
 *
 *              Use in source and binary forms, and redistribution in binary
 *              forms, without modification, are permitted provided that the
 *              following conditions are met:
 *
 *              (i) Redistributions of source code is prohibited without
 *              specific written consent from the original supplier.
 *              Redistributions of source code must retain the above copyright
 *              notice and contact information, this list of conditions and the
 *              following disclaimer.
 *
 *              (ii) Redistribution in binary form is subject to any
 *              non-redistribution stipulations provided by Synesis Software Pty Ltd, or passed
 *              on by an intermediate supplier, at the time of distribution.
 *              Specifically, if a document stipulating one or more binary
 *              components are NOT to be redistributed is obtained from the
 *              supplier at the time of distribution then the user must (a) not
 *              redistribute any material named in that document, and (b) must
 *              redistribute that document with any other materials
 *              redistributed.
 *
 *              Redistributions in binary form must reproduce the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer in the documentation
 *              and/or other materials provided with the distribution.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software (Pty) Ltd.
 *
 *              (iv) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (v) Neither the name of Synesis Software Pty Ltd nor the names of any
 *              subdivisions, employees or agents of Synesis Software Pty Ltd, nor the names of
 *              any other contributors to this software may be used to endorse
 *              or promote products derived from this software without specific
 *              prior written permission.
 *
 *              (vi) Prominent citations stating the name of the Synesis Software Pty Ltd are
 *              placed into the source-code, executable programs and any
 *              associated literature of any work based on or using the
 *              software. Citations in the executables programs must be in the
 *              About Box (or similar concept) of a program interacting with the
 *              user via a Graphical User Interface (GUI), or in the form of an
 *              information string to be printed to the console along with other
 *              help and version information of a non-GUI or command-line
 *              program.
 *
 *              This source code is provided by Synesis Software Pty Ltd "as is" and any
 *               warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software Pty Ltd be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#include <MLBase.h>

#include "MLTssStr.h"

#include <MLAssert.h>
#include <MLCSPriv.h>
#include <MLTrace.h>
#include <MLThread.h>
#include <MLBfWrap.h>
#include <MLSmpCnt.h>
#include <MLMemory.h>
#include <MLSmpVct.h>
#include <MLSmpLst.h>

#include <stlsoft.h>
#include <stlsoft/union_cast.hpp>
#include <stlsoft_simple_type_converter.h>

#ifdef __SYNSOFT_DBS_API_PTHREADS
# include <pthread.h>
#endif /* __SYNSOFT_DBS_PTHREADS */

#ifndef __SYNSOFT_DBS_MULTITHREADED_SUPPORT
# error The TSS Store API is not compatible with non-multithreaded compilations
#endif /* !__SYNSOFT_DBS_MULTITHREADED_SUPPORT */

#ifdef __SYNSOFT_DBS_API_WIN32
# include <MWBAse.h>
#endif /* __SYNSOFT_DBS_API_WIN32 */

//#define _USE_STD_VECTOR

//#define _USE_TLS_HACK

#ifdef __SYNSOFT_DBS_COMPILER_SUPPORTS_PRAGMA_MESSAGE
# pragma message(_sscomp_fileline_message("This needs to be rewritten so that the API lock is skipped."))
# pragma message(_sscomp_fileline_message("  Do it by allocating fixed sized key pages (containing ptrs only)"))
# pragma message(_sscomp_fileline_message("  and the end of each key page is a pointer to the next. Keys are"))
# pragma message(_sscomp_fileline_message("  never reclaimed, so that the access to the pages is via atomic ops only"))
# pragma message(_sscomp_fileline_message("  and the access to the creation of each new key page is via double-checked locking"))
#endif /* __SYNSOFT_DBS_COMPILER_SUPPORTS_PRAGMA_MESSAGE */


#ifdef _USE_TLS_HACK
# include <MLAtomic.h>
#endif /* _USE_TLS_HACK */

#ifdef _USE_STD_VECTOR
# include <vector>
#endif /* _USE_STD_VECTOR */

////////////////////////////////////////////////////////////////////////////////

__SYNSOFT_GEN_NS_START(SynesisStd)
__SYNSOFT_GEN_USING(::SynesisDev::Scope_CoreSync)
__SYNSOFT_GEN_USING(::SynesisDev::CoreSync_Init)
__SYNSOFT_GEN_USING(::SynesisDev::CoreSync_Uninit)

////////////////////////////////////////////////////////////////////////////////
// Typedefs

typedef	ThreadId					threadid_t;

struct ThreadEntry
{
	threadid_t	thread;
	TssValue	datum;
};

#ifdef _USE_STD_VECTOR
 typedef ::std::vector<ThreadEntry>	TssDataBuffer;
 typedef ::std::vector<ThreadId>	ThreadIdBuffer;
#else
 typedef BufferWrap<ThreadEntry>	TssDataBuffer;
 typedef BufferWrap<ThreadId>		ThreadIdBuffer;
#endif /* _USE_STD_VECTOR */

__SyPtrType1(TssDataBuffer)

// Contains the handle, the functions, and the vector of TssThreadValue.
struct TssKey_
{
	HTssKey						hKey;			// the handle of this entry
	PFnTssSlotClose				pfnClose;		// the function to close the entry
	PFnTssKeyClientDisconnnect	pfnDisconnect;	// the function to disconnect the containing library
	Boolean						bCloseOnAssign;	// flag to determine whether to execute the close function on assignment
	TssDataBuffer				threadData;		// SyBuffer containing the data

	TssKey_(	HTssKey						hKey
			,	PFnTssSlotClose				_pfnClose
			,	PFnTssKeyClientDisconnnect	_pfnDisconnect
			,	Boolean						_bCloseOnAssign)
		: hKey(hKey)
		, pfnClose(_pfnClose)
		, pfnDisconnect(_pfnDisconnect)
		, bCloseOnAssign(_bCloseOnAssign)
	{}
	TssKey_(TssKey_ const &rhs)
		: hKey(rhs.hKey)
		, pfnClose(rhs.pfnClose)
		, pfnDisconnect(rhs.pfnDisconnect)
		, bCloseOnAssign(rhs.bCloseOnAssign)
		, threadData(rhs.threadData)
	{}

	_mP(void) operator new(Size /* si */, LPVoid pv)
	{
		return pv;
	}
};
__SyPtrType1(TssKey_)

// class TssImpl
//
// This class stores all the (notional) Tss objects, in a simple vector
// of 
class TssImpl
{
public:
	typedef	TssImpl	Class;

// Construction
public:
	TssImpl();
	~TssImpl() SyThrow_None();

// Entries
public:
	HTssKey	CreateEntry(PFnTssSlotClose				pfnClose
					,	PFnTssKeyClientConnnect		pfnConnect
					,	PFnTssKeyClientDisconnnect	pfnDisconnect
					,	Boolean						bCloseOnAssign);
	void	CloseEntry(	HTssKey						hKey);

// Attach/Detach thread
public:
	void	AttachThread();
	void	DetachThread();
	Bool	HasThreadAttached() const;

// Thread data
public:
	void		SetSlotValue(HTssKey hKey, TssValue pv, PTssValue ppv);
	TssValue	GetSlotValue(HTssKey hKey) const;

public:
	_mP(void) operator new(Size si, LPVoid pv);
	void operator delete(LPVoid pv) SyThrow_None();
#ifdef __SYNSOFT_DBS_PLACEMENT_DELETE_SUPPORT
	void operator delete(LPVoid , LPVoid pv) SyThrow_None();
#endif // __SYNSOFT_DBS_PLACEMENT_DELETE_SUPPORT

// Implementation
private:
	PThreadId	FindOrCreateThread(HTssKey hKey, Bool bCreate);
	Int			GetThreadIndex(ThreadId thread) const;

#ifdef __SYNSOFT_DBS_API_PTHREADS
	static void __SYNSOFT_MOD_CALLCONV_CDECL PThreads_key_destructor(void *);
#endif /* __SYNSOFT_DBS_PTHREADS */

// Members
private:
	HSmpCont_ptr	m_scEntries;	// Simple container of TssKey_.
	ThreadIdBuffer	m_threads;		// Buffer_ of thread ids. Doesn't matter if repeated, since will 
	HTssKey			m_hStore;		// Next store.
#ifdef __SYNSOFT_DBS_API_PTHREADS
	static Boolean	sm_Initialised;
	pthread_key_t	m_pthreadsKey;	// Key used to hook into thread cleanup
#endif /* __SYNSOFT_DBS_PTHREADS */

_SY_PRIVATE_COPYCTOR_AND_OPEQUALS2(TssImpl, Class)
};

__SyPtrType1(TssImpl)

inline LPVoid TssImpl::operator new(Size /*si*/, LPVoid pv)
{
	return pv;
}

inline void TssImpl::operator delete(LPVoid /*pv*/) SyThrow_None()
{}

#ifdef __SYNSOFT_DBS_PLACEMENT_DELETE_SUPPORT
inline void TssImpl::operator delete(LPVoid , LPVoid /* pv */) SyThrow_None()
{}
#endif // __SYNSOFT_DBS_PLACEMENT_DELETE_SUPPORT

#ifdef _USE_TLS_HACK
static DWORD	s_hackIndex;
#endif /* _USE_TLS_HACK */


static TssImpl &_sg_tsslib(Int iInit)
{
	struct TssImplStorage
	{
		Byte bytes[sizeof(TssImpl)];
	};
	static TssImplStorage	s_storage;
	static PTssImpl			s_ptss	=	NULL;
	static Long				s_cRefs		=	0;

	SyAssert(s_cRefs >= 0);

	if(	iInit > 0 && 
		++s_cRefs < 2)
	{
		// Create the tss store
		SyAssert(!s_ptss);
		s_ptss = new(&s_storage) TssImpl;

#ifdef _USE_TLS_HACK
		s_hackIndex = ::TlsAlloc();
#endif /* _USE_TLS_HACK */
	}
	else if(iInit < 0 && 
			--s_cRefs == 0)
	{
#ifdef _USE_TLS_HACK
		::TlsFree(s_hackIndex);
#endif /* _USE_TLS_HACK */

		// Create the tss store
		SyAssert((LPVoid)s_ptss == &s_storage);
		DeleteSetNull(s_ptss);
	}

	return *s_ptss;
}

#define	sg_tssstore	_sg_tsslib(0)
#define	sg_cs		CoreSyncCS(Tss)

////////////////////////////////////////////////////////////////////////////////
// API implementation

SYFNDECL(SInt32) Tss_Init(void)
{
	if(CoreSync_Init() < 0)
	{
		return -1;
	}
	else if(Mem_Init() < 0)
	{
		CoreSync_Uninit();
		return -1;
	}
	else if(SimpleContainer_Init() < 0)
	{
		Mem_Uninit();
		CoreSync_Uninit();

		return -1;
	}

	Scope_CoreSync	lock(sg_cs);

	_sg_tsslib(1);	// Create the store

	return 0;
}

SYFNDECL(void) Tss_Uninit(void)
{
	{
		Scope_CoreSync	lock(sg_cs);

		_sg_tsslib(-1);	// Close the store
	}

	SimpleContainer_Uninit();
	Mem_Uninit();
	CoreSync_Uninit();
}

SYFNDECL(HTssKey) Tss_CreateKey(PFnTssSlotClose				pfnClose
							,	PFnTssKeyClientConnnect		pfnConnect
							,	PFnTssKeyClientDisconnnect	pfnDisconnect
							,	Boolean						bCloseOnAssign)
{
	Scope_CoreSync	lock(sg_cs);

	return sg_tssstore.CreateEntry(pfnClose, pfnConnect, pfnDisconnect, bCloseOnAssign);
}

SYFNDECL(void) Tss_CloseKey(HTssKey hKey)
{
	Scope_CoreSync	lock(sg_cs);

	// Close an entry.
	sg_tssstore.CloseEntry(hKey);
}

SYFNDECL(void) Tss_SetSlotValue(HTssKey hKey, TssValue data, PTssValue pdatum)
{
#ifdef _USE_TLS_HACK
	Atomic_Increment((long&)s_hackIndex);
	Atomic_Decrement((long&)s_hackIndex);
	::TlsSetValue(s_hackIndex, data);
#else /* ? _USE_TLS_HACK */
	Scope_CoreSync	lock(sg_cs);

	// Set thread datum.
	sg_tssstore.SetSlotValue(hKey, data, pdatum);
#endif /* _USE_TLS_HACK */
}

SYFNDECL(TssValue) Tss_GetSlotValue(HTssKey hKey)
{
#ifdef _USE_TLS_HACK
	Atomic_Increment((long&)s_hackIndex);
	Atomic_Decrement((long&)s_hackIndex);
	return ::TlsGetValue(s_hackIndex);
#else /* ? _USE_TLS_HACK */
	Scope_CoreSync	lock(sg_cs);

	// Set thread datum.
	return sg_tssstore.GetSlotValue(hKey);
#endif /* _USE_TLS_HACK */
}

#ifdef __INTEL_COMPILER
# pragma warning(disable : 1418)
#endif /* __INTEL_COMPILER */

SYFNDECL(void) Tss_ThreadAttach(void)
{
	Scope_CoreSync	lock(sg_cs);

	// Thread attach.
	sg_tssstore.AttachThread();
}

SYFNDECL(void) Tss_ThreadDetach(void)
{
	Scope_CoreSync	lock(sg_cs);

	// Thread detach.
	sg_tssstore.DetachThread();
}

////////////////////////////////////////////////////////////////////////////////
// Container process & query callbacks

static void _CloseEntry_(	RTssValue				rdatum,
							PFnTssSlotClose			pfnClose,
#ifdef __SYNSOFT_DBS_DEBUG
							Boolean					bDumpLeak,
#else
#define	_CloseEntry_(_rdata, _pfnClose, _bDumpLeak, _thread) \
		_CloseEntry_(_rdata, _pfnClose, _thread)
#endif // __SYNSOFT_DBS_DEBUG
							ThreadId				thread)
{
	STLSOFT_SUPPRESS_UNUSED(thread);

	if(TssValue_Null != rdatum)
	{
		SyTraceA2("Tss Store information: un-closed resource 0x%08x for thread id 0x%08x; closing now.\n", rdatum, thread);

		TssValue	datum	=	rdatum;

		rdatum = TssValue_Null;

		if(pfnClose != NULL)
		{
			(*pfnClose)(datum);
		}
	}
}

// This procedure checks that a given entry's data item for a given thread has
// been released, and, if not, it releases it and traces the leak. pvParam is a
// pointer to a ThreadIndexInfo structure.
static Boolean __SYNSOFT_MOD_CALLCONV_DEFAULT ProcessFnCloseEntry(LPVoid pvItem, LPVoid /* pvParam */, Size /*siItem*/)
{
	// pvItem is the address of the pointer to the store
	PTssKey_				pentry	=	(PTssKey_)pvItem;
	TssDataBuffer::iterator	b		=	pentry->threadData.begin();
	TssDataBuffer::iterator	e		=	pentry->threadData.end();

	// Check iItem is within range
	for(; b != e; ++b)
	{
		_CloseEntry_((*b).datum, pentry->pfnClose, true, (*b).thread);
	}

	return true;
}

// This procedure disconnects a given item, and empties out its data. pvParam is
// not used.
static Boolean __SYNSOFT_MOD_CALLCONV_DEFAULT ProcessFnDisconnectEntry(LPVoid pvItem, LPVoid /*pvParam*/, Size /*siItem*/)
{
	PTssKey_	pentry = (PTssKey_)pvItem;

	if(pentry->pfnDisconnect != NULL)
	{
		(*pentry->pfnDisconnect)();
	}

	return true;
}

// This function inserts a thread space into the entry's data. pvParam points to
// a ThreadIndexInfo structure containing the thread id and the new index to be 
// inserted.
static Boolean __SYNSOFT_MOD_CALLCONV_DEFAULT ProcessFnInsertAttachThread(LPVoid pvItem, LPVoid pvParam, Size /*siItem*/)
{
	// pvItem is the address of the pointer to the store
	threadid_t	thread	=	(threadid_t)pvParam;
	PTssKey_	pentry	=	(PTssKey_)pvItem;

	ThreadEntry	te		=	{ thread, TssValue_Null };

#ifdef _DEBUG
	{
		TssDataBuffer::iterator	b		=	pentry->threadData.begin();
		TssDataBuffer::iterator	e		=	pentry->threadData.end();

		for(; b != e; ++b)
		{
			SyAssert((*b).thread != thread);
		}
	}
#endif /* _DEBUG */

	// Re-allocate buffer to insert item, move the existing up one, and insert the item
	pentry->threadData.push_back(te);

	return true;
}

// This function removes a thread space from the entry's data. pvParam points to 
// a ThreadIndexInfo structure containing the thread id and the index to be 
// removed. 
static Boolean __SYNSOFT_MOD_CALLCONV_DEFAULT ProcessFnDetachThread(LPVoid pvItem, LPVoid pvParam, Size /*siItem*/)
{
	// pvItem is the address of the pointer to the store
	PTssKey_				pentry	=	(PTssKey_)pvItem;
	threadid_t				thread	=	(threadid_t)pvParam;
	TssDataBuffer::iterator	b		=	pentry->threadData.begin();
	TssDataBuffer::iterator	e		=	pentry->threadData.end();

	// Check iItem is within range
	for(; b != e; ++b)
	{
		if((*b).thread == thread)
		{
			_CloseEntry_((*b).datum, pentry->pfnClose, false, (*b).thread);

			pentry->threadData.erase(b);

			break;
		}
	}

	return true;
}

// This procedure is used to find an entry by the store handle. pvParam is the
// store handle.
static Int __SYNSOFT_MOD_CALLCONV_DEFAULT QueryFnFindEntryByHEntry(LPCVoid pvItem, LPVoid pvParam, Size /*siItem*/)
{
	// pvItem is the address of the pointer to the store
	PTssKey_	pentry = (PTssKey_)pvItem;
	HTssKey		hKey = (HTssKey)pvParam;

	return pentry->hKey == hKey;
}

// This procedure copies one TssKey_ to another. pcvSrc points to the source item.
static void __SYNSOFT_MOD_CALLCONV_DEFAULT CopyFnTssEntry(LPVoid pvDest, LPCVoid pcvSrc, Size /*siItem*/)
{
	// pvItem is the address of the pointer to the store
	PTssKey_	pentryDest	=	(PTssKey_)pvDest;
	PTssKey_	pcentrySrc	=	(PTssKey_)pcvSrc;

	// Copy the elements
	new(pentryDest) TssKey_(*pcentrySrc);
}

// This procedure erases a TssKey_. It does not report on any leaks.
static Boolean __SYNSOFT_MOD_CALLCONV_DEFAULT EraseFnTssEntry(LPVoid pvItem, Size /* siItem */)
{
	// pvItem is the address of the pointer to the store
	PTssKey_	pentry = (PTssKey_)pvItem;

	pentry->~TssKey_();

	return false;
}

////////////////////////////////////////////////////////////////////////////////
// TssImpl implementation

#ifdef __SYNSOFT_DBS_API_PTHREADS
/* static */ Boolean TssImpl::sm_Initialised	=	false;
#endif /* __SYNSOFT_DBS_PTHREADS */

#ifdef __SYNSOFT_DBS_API_PTHREADS
/* static */ void __SYNSOFT_MOD_CALLCONV_CDECL TssImpl::PThreads_key_destructor(void * /* pv */)
{
	if(sm_Initialised) // This is needed to avoid a callback from within the main thread, which has already closed down the API
	{
		Tss_ThreadDetach();
	}
}
#endif /* __SYNSOFT_DBS_PTHREADS */

TssImpl::TssImpl()
#if 1
	: m_scEntries(SimpleVector_Create(	NULL,
#else
	: m_scEntries(SimpleList_Create(	NULL,
#endif /* 0 */
										sizeof(TssKey_),
										CopyFnTssEntry,
										EraseFnTssEntry,
										NULL,
										NULL,
										NULL))
	, m_hStore(NULL)
{
#ifdef __SYNSOFT_DBS_API_PTHREADS
	pthread_key_create(&m_pthreadsKey, PThreads_key_destructor);

	sm_Initialised = true;
#endif /* __SYNSOFT_DBS_PTHREADS */

	// Attach the initial thread
	AttachThread();
}

TssImpl::~TssImpl() SyThrow_None()
{
	// Now we search all the entries, for each thread value, and report any leaks
	SimpleContainer_Process(m_scEntries, NULL, NULL, ProcessFnCloseEntry, NULL);

	// Now close all the entries.
	SimpleContainer_Process(m_scEntries, NULL, NULL, ProcessFnDisconnectEntry, NULL);

#ifdef __SYNSOFT_DBS_API_PTHREADS
	sm_Initialised = false;
#endif /* __SYNSOFT_DBS_PTHREADS */
}

HTssKey TssImpl::CreateEntry(	PFnTssSlotClose				pfnClose
							,	PFnTssKeyClientConnnect		pfnConnect
							,	PFnTssKeyClientDisconnnect	pfnDisconnect
							,	Boolean						bCloseOnAssign)
{
	// Create an entry.

	HTssKey		hKey;
	TssKey_		entry(	++m_hStore,
						pfnClose,
						pfnDisconnect,
						bCloseOnAssign);
	PTssKey_	pentry = (PTssKey_)SimpleContainer_AddItem(m_scEntries, &entry);

	if(SimpleContainer_IsEndOrError(m_scEntries, pentry))
	{
		hKey = NULL;
	}
	else
	{
		// Insert all the current threads into this entry
		ThreadIdBuffer::const_iterator	b	=	m_threads.begin();
		ThreadIdBuffer::const_iterator	e	=	m_threads.end();

		for(; b != e; ++b)
		{
			ThreadEntry	te = { *b, TssValue_Null };

			pentry->threadData.push_back(te);
		}

		// Connect to the client API
		if(NULL != pfnConnect)
		{
			pfnConnect();
		}

		hKey = entry.hKey;
	}

	return hKey;
}

void TssImpl::CloseEntry(HTssKey hKey)
{
	PTssKey_	pentry = (PTssKey_)SimpleContainer_Query(m_scEntries, NULL, NULL, QueryFnFindEntryByHEntry, (LPVoid)hKey);

	if(NULL == pentry)
	{
		SyTraceA1("Tss Store error: Entry (%08x) not recognised!\n", hKey);
		SyAssert(0);
	}
	else
	{
		TssDataBuffer::iterator	b	=	pentry->threadData.begin();
		TssDataBuffer::iterator	e	=	pentry->threadData.end();

		// Check iItem is within range
		for(; b != e; ++b)
		{
			_CloseEntry_((*b).datum, pentry->pfnClose, true, (*b).thread);
		}

		// Disconnect the entry
		ProcessFnDisconnectEntry(pentry, NULL, sizeof(TssKey_));

		// Now remove the store entry
		SimpleContainer_RemoveItem(m_scEntries, pentry);
	}
}

void TssImpl::SetSlotValue(HTssKey hKey, TssValue data, PTssValue pdatum)
{
	// Get the current thread id, and determine the index into the data vectors
	ThreadId	thread	=	GetActiveThreadId();
	Long		index	=	GetThreadIndex(thread);

	if(index < 0)
	{
		AttachThread();
		index = GetThreadIndex(thread);
		SyAssert(index >= 0);
	}

	// Now find the entry
	PTssKey_	pentry = (PTssKey_)SimpleContainer_Query(m_scEntries, NULL, NULL, QueryFnFindEntryByHEntry, (LPVoid)hKey);

	// Must exist
	SyAssert(!SimpleContainer_IsEndOrError(m_scEntries, pentry));

	TssDataBuffer::iterator	b		=	pentry->threadData.begin();
	TssDataBuffer::iterator	e		=	pentry->threadData.end();

	// Check iItem is within range
	for(; b != e; ++b)
	{
		if((*b).thread == thread)
		{
			RTssValue	rdatum = (*b).datum;

			if(data != rdatum)
			{
				if(pdatum != NULL)
				{
					*pdatum = rdatum;
				}
				else if(TssValue_Null != rdatum &&
						pentry->bCloseOnAssign)
				{
					pentry->pfnClose(rdatum);
				}

				rdatum = data;
			}

			break;
		}
	}
}

TssValue TssImpl::GetSlotValue(HTssKey hKey) const 
{
	ThreadId	thread	=	GetActiveThreadId();
	Long		index	=	GetThreadIndex(thread);

	if(index < 0)
	{
#ifdef __SYNSOFT_DBS_API_PTHREADS
		SyTraceA2("Tss Store informational: Cannot get thread datum. This threads (%08x) is not yet attached to store (%08x)!\n", thread, hKey);
#else
		SyTraceA2("Tss Store warning: Cannot get thread datum. Thread (%08x) not attached to store (%08x)!\n", thread, hKey);
		SyAssert(0);
#endif /* __SYNSOFT_DBS_PTHREADS */
		return NULL;
	}

	PCTssKey_	pcentry = (PCTssKey_)SimpleContainer_Query(m_scEntries, NULL, NULL, QueryFnFindEntryByHEntry, (LPVoid)hKey);

	SyAssert(!SimpleContainer_IsEndOrError(m_scEntries, pcentry));

	TssDataBuffer::const_iterator	b	=	pcentry->threadData.begin();
	TssDataBuffer::const_iterator	e	=	pcentry->threadData.end();

	// Check iItem is within range
	for(; b != e; ++b)
	{
		if((*b).thread == thread)
		{
			return (*b).datum;
		}
	}

	SyAssert(0);

	return TssValue_Null;
}

void TssImpl::AttachThread()
{
	// Simply iterate through the thread array to find the insertion point, then
	// insert the thread into the thread array. Then process all entries with the
	// new thread id and index

	// Determine the insertion point in the thread array
	ThreadId					thread	=	GetActiveThreadId();
	ThreadIdBuffer::iterator	b		=	m_threads.begin();
	ThreadIdBuffer::iterator	e		=	m_threads.end();

	for(; b != e; ++b)
	{
		SyAssert(0 != (*b));

		if(thread == (*b))
		{
			return;	// Already in the array
		}

		if(thread < (*b))
		{
			break;	// Found insertion point
		}
	}

	// Insert the new thread
	m_threads.insert(b, thread);

	// Make all entries aware of new thread
	SimpleContainer_Process(m_scEntries, NULL, NULL, ProcessFnInsertAttachThread, stlsoft_ns_qual(make_union_cast)<LPVoid>(thread));

#ifdef __SYNSOFT_DBS_API_PTHREADS
	// Make sure we get a callback
	pthread_setspecific(m_pthreadsKey, reinterpret_cast<void*>(1));
#endif /* __SYNSOFT_DBS_PTHREADS */
}

void TssImpl::DetachThread()
{
	// Process all entries to remove the thread data, then remove from the 
	// thread array.

	// Determine the removal point from the thread array
	ThreadId	thread	=	GetActiveThreadId();
	Long		i		=	GetThreadIndex(thread);

	if(i < 0)
	{
		SyTraceA1("Unrecognised thread (%08x) detaching\n", thread);
	}
	else
	{
		// Thread was recognised, i.e. created since the library was 
		// initialised

		// Remove new data item (NULL) into entries
		SimpleContainer_Process(m_scEntries, NULL, NULL, ProcessFnDetachThread, stlsoft_ns_qual(make_union_cast)<LPVoid>(thread));

		// Remove the thread
#ifdef _USE_STD_VECTOR
		m_threads.erase(&m_threads[i]);
#else
		m_threads.RemoveItem(i);
#endif /* _USE_STD_VECTOR */
	}
}

Bool TssImpl::HasThreadAttached() const
{
	// Determine the removal point from the thread array
	ThreadId	thread = GetActiveThreadId();

	return GetThreadIndex(thread) >= 0;
}

Int TssImpl::GetThreadIndex(ThreadId thread) const
{
	// Determine the removal point from the thread array
	PCThreadId	pthread;
	ULong		cThreads = m_threads.size();
	ULong		i;
	Int			iRet;

	for(pthread = &m_threads[0], i = 0, iRet = -1; i < cThreads; ++pthread, ++i)
	{
		SyAssert(*pthread != 0);						// The thread entry cannot be 0
		SyAssert(i == 0 || *pthread > *(pthread-1));	// The threads must be ordered

		if(thread == *pthread)
		{
			iRet = i;
			break;
		}
#if 0
		else if(thread < *pthread)
		{
			break;
		}
#endif /* 0 */
	}

	return iRet;
}

////////////////////////////////////////////////////////////////////////////////

__SYNSOFT_GEN_NS_END_IMPL(SynesisStd)

////////////////////////////////////////////////////////////////////////////////
